from .harm import HarmCategory, HarmBlockThreshold, HarmProbability
from .part import TextPart, FilePart, ImagePart, VideoPart
from .request import (
    ChatMessage,
    ChatHistory,
    SafetySettings,
    GenerationRequest,
    GenerationRequestParts,
    GenerationConfig,
)
from .response import GenerationCandidate, GenerationResponse
from .upload import UploadFile, UploadedFile



import mimetypes

from pydantic import Field, HttpUrl

from .base import BaseModel
from .part import FilePart, FilePartData


class UploadFile(BaseModel):
    file_path: str = Field(..., description="Path to the file.")

    mime_type: str | None = Field(None, description="MIME type of the file.")

    body: dict | None = Field(None, description="The body of the request.")

    @classmethod
    def from_path(cls, file_path: str, body: dict | None = None) -> "UploadFile":
        mine_type, _ = mimetypes.guess_type(file_path)
        return cls(file_path=file_path, mime_type=mine_type, body=body)


class UploadedFile(BaseModel):
    name: str = Field(..., description="Unique identifier for the file.")

    display_name: str | None = Field(
        None, alias="displayName", description="Human-readable name for the file."
    )

    mime_type: str | None = Field(..., alias="mimeType", description="MIME type of the file.")

    size_bytes: int = Field(..., alias="sizeBytes", description="Size of the file in bytes.")

    create_time: str = Field(..., alias="createTime", description="Time the file was created.")

    update_time: str = Field(..., alias="updateTime", description="Time the file was updated.")

    expiration_time: str | None = Field(
        None, alias="expirationTime", description="Time the file will expire."
    )

    sha256_hash: str = Field(..., alias="sha256Hash", description="SHA-256 hash of the file.")

    uri: HttpUrl = Field(..., description="URI of the file.")

    def to_file_part(self):
        return FilePart(file_data=FilePartData(file_uri=str(self.uri), mime_type=self.mime_type))



from enum import Enum


class HarmCategory(str, Enum):
    SEXUALLY_EXPLICIT = "HARM_CATEGORY_SEXUALLY_EXPLICIT"
    HATE_SPEECH = "HARM_CATEGORY_HATE_SPEECH"
    HARASSMENT = "HARM_CATEGORY_HARASSMENT"
    DANGEROUS_CONTENT = "HARM_CATEGORY_DANGEROUS_CONTENT"


class HarmBlockThreshold(str, Enum):
    HARM_BLOCK_THRESHOLD_UNSPECIFIED = "HARM_BLOCK_THRESHOLD_UNSPECIFIED"
    BLOCK_LOW_AND_ABOVE = "BLOCK_LOW_AND_ABOVE"         # Content with NEGLIGIBLE will be allowed.
    BLOCK_MEDIUM_AND_ABOVE = "BLOCK_MEDIUM_AND_ABOVE"   # Content with NEGLIGIBLE and LOW will be allowed.
    BLOCK_ONLY_HIGH = "BLOCK_ONLY_HIGH"                 # Content with NEGLIGIBLE, LOW, and MEDIUM will be allowed.
    BLOCK_NONE = "BLOCK_NONE"                           # All content will be allowed.


class HarmProbability(str, Enum):
    NEGLIGIBLE = "NEGLIGIBLE"   # Content has a negligible chance of being unsafe.
    LOW = "LOW"                 # Content has a low chance of being unsafe.
    MEDIUM = "MEDIUM"           # Content has a medium chance of being unsafe.
    HIGH = "HIGH"               # Content has a high chance of being unsafe.



import typing

from pydantic import Field

from .base import BaseModel
from .harm import HarmCategory, HarmBlockThreshold
from .part import TextPart, FilePart


class ChatMessage(BaseModel):
    role: typing.Literal["model", "user"] = Field(
        ..., description="Role of the author of the parts."
    )
    parts: list[TextPart | FilePart] = Field(
        ..., description="Parts of the chat message."
    )


class ChatHistory(BaseModel):
    messages: list[ChatMessage] = Field(
        ..., description="History of chat messages."
    )


class GenerationRequestParts(BaseModel):
    parts: list[TextPart | FilePart] = Field(
        ..., description="Parts of the request."
    )


class GenerationConfig(BaseModel):
    stop_sequences: list[str] | None = Field(
        None,
        alias="stopSequences",
        description="The set of character sequences (up to 5) that will stop output generation.",
    )

    candidate_count: int | None = Field(
        None,
        alias="candidateCount",
        description=(
            "Number of generated responses to return. This value must be between [1, 8], inclusive. "
            "If unset, this will default to 1."
        ),
    )

    max_output_tokens: int | None = Field(
        None,
        alias="maxOutputTokens",
        description=(
            "The maximum number of tokens to generate. The default value varies by model, see the "
            "Model.output_token_limit attribute of the Model returned from the `get_model` function."
        ),
    )

    temperature: float | None = Field(
        None,
        description=(
            "Controls randomness in generation. Lower values make the model more deterministic. "
            "High values make the model more creative."
        ),
    )

    top_p: float | None = Field(
        None,
        alias="topP",
        description="The maximum cumulative probability of tokens to consider when sampling.",
    )

    top_k: int | None = Field(
        None,
        alias="topK",
        description="The maximum number of tokens to consider when sampling.",
    )


class SafetySettings(BaseModel):
    category: HarmCategory | None = Field(
        None, description="The category of harmful content to block."
    )

    threshold: HarmBlockThreshold | None = Field(
        None, description="The threshold of harmful content to block."
    )


class GenerationRequest(BaseModel):
    contents: list[GenerationRequestParts | ChatMessage] | dict = Field(
        ..., description="Contents of the request."
    )

    generation_config: GenerationConfig | dict | None = Field(
        None, alias="generationConfig", description="Generation configuration."
    )

    safety_settings: SafetySettings | dict | None = Field(
        None, alias="safetySettings", description="Safety settings."
    )



import os
from pathlib import Path

import requests
import googleapiclient.discovery as g_discovery

from .chat import ChatSession
from .schemas import (
    ChatMessage,
    ChatHistory,
    GenerationConfig,
    GenerationRequest,
    GenerationRequestParts,
    GenerationResponse,
    SafetySettings,
    TextPart,
    ImagePart,
    FilePart,
    VideoPart,
    UploadFile,
    UploadedFile,
)


class GeminiClient:
    def __init__(
        self,
        api_key: str | None = None,
        version: str = "v1beta",
    ):
        api_key = api_key or os.getenv("GEMINI_API_KEY")

        if api_key is None:
            raise ValueError("Gemini API (`GEMINI_API_KEY`) key must be provided")

        rsp = requests.get(
            "https://generativelanguage.googleapis.com/$discovery/rest",
            params={"version": version, "key": api_key},
        )
        rsp.raise_for_status()

        self.genai_service = g_discovery.build_from_document(
            rsp.content, developerKey=api_key
        )

    @staticmethod
    def normalize_prompt(prompt: list | str) -> list:
        parts = []

        if isinstance(prompt, str):
            prompt = [prompt]

        for part in prompt:
            if isinstance(part, str):
                parts.append(TextPart(text=f"\n{part}\n"))  # Take care of the newline more elegantly
            elif isinstance(part, (FilePart, ImagePart)):
                parts.append(part)
            elif isinstance(part, VideoPart):
                parts.extend(part.content_parts())
            else:
                raise ValueError(f"Invalid prompt part type: {type(part)} ({part})")

        return parts

    def generate(
        self,
        model: str,
        prompt: GenerationRequest | ChatHistory | list | str,
        generation_config: GenerationConfig | dict | None = None,
        safety_settings: SafetySettings | dict | None = None,
    ) -> GenerationResponse:
        generation_request = self._prepare_generation_request(
            prompt,
            generation_config=generation_config,
            safety_settings=safety_settings,
        )

        rsp = (
            self.genai_service
                .models()
                .generateContent(
                    model=model,
                    body=generation_request.model_dump(by_alias=True, exclude_none=True),
                )
                .execute()
        )

        return GenerationResponse.model_validate(rsp)

    def start_chat(
        self,
        model: str,
        history: list[ChatMessage] | ChatHistory | None = None,
        generation_config: GenerationConfig | dict | None = None,
        safety_settings: SafetySettings | dict | None = None,
    ) -> ChatSession:
        if isinstance(history, ChatHistory):
            history = history.messages

        return ChatSession(
            self,
            model,
            history=history,
            generation_config=generation_config,
            safety_settings=safety_settings,
        )

    def _prepare_generation_request(
        self,
        prompt: GenerationRequest | ChatHistory | list | str,
        generation_config: GenerationConfig | dict | None = None,
        safety_settings: SafetySettings | dict | None = None,
    ) -> GenerationRequest:
        if isinstance(prompt, GenerationRequest):
            request = prompt
        elif isinstance(prompt, ChatHistory):
            request = GenerationRequest(contents=prompt.messages)
        else:
            parts = self.normalize_prompt(prompt)
            request = GenerationRequest(
                contents=[GenerationRequestParts(parts=parts)]
            )

        if generation_config is not None:
            if not isinstance(generation_config, GenerationConfig):
                generation_config = GenerationConfig.model_validate(generation_config)
            request.generation_config = generation_config

        if safety_settings is not None:
            if not isinstance(safety_settings, SafetySettings):
                safety_settings = SafetySettings.model_validate(safety_settings)
            request.safety_settings = safety_settings

        return request

    def upload_image(self, image_path: str) -> ImagePart:
        if not Path(image_path).exists():
            raise FileNotFoundError(f"Image file not found: {image_path}")

        uploaded_file = self._upload_file(UploadFile.from_path(image_path))

        return uploaded_file.to_file_part()

    def upload_video(self, video_path: str) -> VideoPart:
        if not Path(video_path).exists():
            raise FileNotFoundError(f"Video file not found: {video_path}")

        raise NotImplementedError("Video upload is not yet implemented")

    def _upload_file(self, file: UploadFile) -> UploadedFile:
        rsp = (
            self.genai_service.media()
            .upload(
                media_body=file.file_path,
                media_mime_type=file.mime_type,
                body=file.body,
            )
            .execute()
        )

        return UploadedFile.model_validate(rsp["file"])

    def _upload_files(self, *files: list[UploadFile]) -> list[UploadedFile]:
        return [self._upload_file(file) for file in files]



from pydantic import Field

from .base import BaseModel


class TextPart(BaseModel):
    text: str = Field(..., description="Text content.")


class FilePartData(BaseModel):
    file_uri: str = Field(..., alias="fileUri", description="URI of the file.")
    mime_type: str = Field(..., alias="mimeType", description="MIME type of the file.")


class FilePart(BaseModel):
    file_data: FilePartData = Field(..., description="URL of the file.")


class ImagePart(FilePart):
    ...


class VideoPart(BaseModel):
    time_spans: list[TextPart] = Field(..., description="Time spans of the video.")

    frames: list[ImagePart | FilePart] = Field(..., description="Frames of the video.")

    def content_parts(self) -> list[TextPart | ImagePart | FilePart]:
        parts = []

        for span, frame in zip(self.time_spans, self.frames):
            parts.append(span)
            parts.append(frame)

        return parts



from pydantic import Field

from .base import BaseModel
from .harm import HarmCategory, HarmProbability
from .part import TextPart


class Content(BaseModel):
    role: str = Field(..., description="Role of the author of the content.")

    parts: list[TextPart] = Field(..., description="Parts of the content.")


class SafetyRating(BaseModel):
    category: HarmCategory = Field(..., description="The category of the content.")

    probability: HarmProbability = Field(
        ..., description="The probability of the content falling into the category."
    )


class GenerationCandidate(BaseModel):
    index: int = Field(..., description="Index of the candidate in the list.")

    content: Content | None = Field(
        None, description="Content generated by the model."
    )

    finish_reason: str | None = Field(
        None, alias="finishReason", description="The reason the generation stopped."
    )

    safety_ratings: list[SafetyRating] | None = Field(
        None, alias="safetyRatings", description="Safety ratings of the content."
    )


class GenerationResponse(BaseModel):
    candidates: list[GenerationCandidate] = Field(
        [], description="The generated responses."
    )



from typing import TYPE_CHECKING

from .schemas import (
    ChatMessage, ChatHistory, GenerationConfig, GenerationResponse, SafetySettings
)


if TYPE_CHECKING:
    from .client import GeminiClient


class ChatSession:
    def __init__(
        self,
        client: "GeminiClient",
        model: str,
        history: list[ChatMessage] | None = None,
        generation_config: GenerationConfig | dict | None = None,
        safety_settings: SafetySettings | dict | None = None,
    ):
        self.client = client
        self.model = model
        self.history = history or []
        self.generation_config = generation_config
        self.safety_settings = safety_settings

    def send_message(
        self,
        message: list | str,
        generation_config: GenerationConfig | dict | None = None,
        safety_settings: SafetySettings | dict | None = None,
    ) -> GenerationResponse:
        parts = self.client.normalize_prompt(message)
        self.history.append(ChatMessage(role="user", parts=parts))

        rsp = self.client.generate(
            self.model,
            ChatHistory(messages=self.history),
            generation_config=generation_config or self.generation_config,
            safety_settings=safety_settings or self.safety_settings,
        )
        rsp_parts = rsp.candidates[0].content.parts
        self.history.append(ChatMessage(role="model", parts=rsp_parts))

        return rsp

    def clear(self):
        self.history = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clear()
